--Libs
local Lume = require("libs.lume")
local Vector = require("libs.vector")

--Other objects
local Arrow = require("classes.arrow")

--A physical object that simulates a bow in a really not accurate way.
local Bow = Class("bow")

Bow.armLength = 35 -- Length of each arm
Bow.armTension = 5000 --The bigger this is the more the bow will want to maintain its original state (not bended)
Bow.lineLength = 88 --The length of the line total (each line will have half of this length)
Bow.lineTension = 30 --The higher the more rigid the line will be (It's a distance joint, this value goes into :setFrequency which makes it act a little more like a spring and then we can have a line that extends)
Bow.divisions = 3 --How many divisions each arm will have (Higher numbers than 6 are really hard to make work and maybe not even possible)
Bow.maxBend = math.rad(90) --Maximum bend of each division
Bow.minBend = math.rad(2) -- Minimum angle of each division.
Bow.handleHeight = 34 -- The size of the handle of the bow
Bow.materialWidth = 2 -- The width of the bow. Probably doesn't has any effect other than changing the mass of the bow (Which actually makes a lot of difference... but it is hard to predict what)


function Bow:initialize()
  self.parts = {}
  self.revoluteJoints = {}
  --The handle
  local handleB = love.physics.newBody(pWorld, 0, 0, "kinematic")
  local handleS = love.physics.newRectangleShape(Bow.materialWidth, Bow.handleHeight)
  local handleF = love.physics.newFixture(handleB, handleS, 1)
  --Just so it doesn't collide with anything
  handleF:setFilterData(0, 0, 0)
  self.parts["handle"] = {body = handleB, shape = handleS, fixture = handleF}
  
  --The arms
  --The lower arm
  local height = Bow.armLength/Bow.divisions
  local lastpartB = handleB
  local offset = Bow.handleHeight/2
  local mul = 1
  self.parts.lower = {}
  self.revoluteJoints.down = {}
  for i = 1, Bow.divisions do
    --Creating the part
    local partB = love.physics.newBody(pWorld, 0, height*(i-.5)+offset, "dynamic")
    local partS = love.physics.newRectangleShape(Bow.materialWidth, height)
    local partF = love.physics.newFixture(partB, partS, .1)
    --Just so it doesn't collide with anything
    partF:setFilterData(0, 0, 0)
    table.insert(self.parts.lower, {body = partB, shape = partS, fixture = partF})
    
    --Joint
    local joint = love.physics.newRevoluteJoint(lastpartB, partB, 0, height*(i-1)+offset, false)
    table.insert(self.revoluteJoints.down, joint)
    --Limits
    joint:setLimitsEnabled(true)
    joint:setLimits(Bow.minBend, Bow.maxBend)
    --Motor
    joint:setMotorEnabled(true)
    joint:setMotorSpeed(-500)
    joint:setMaxMotorTorque(0)
    
    --Set the last part so the next division will be attached to this one
    lastpartB = partB
  end
  local end1 = lastpartB
  
  --The upper arm
  local height = Bow.armLength/Bow.divisions
  local lastpartB = handleB
  local offset = Bow.handleHeight/2
  local mul = 1
  self.parts.upper = {}
  self.revoluteJoints.up = {}
  for i = 1, Bow.divisions do
    --Creating the part
    local partB = love.physics.newBody(pWorld, 0, -height*(i-.5)-offset, "dynamic")
    local partS = love.physics.newRectangleShape(Bow.materialWidth, height)
    local partF = love.physics.newFixture(partB, partS, .1)
    --Just so it doesn't collide with anything
    partF:setFilterData(0, 0, 0)
    table.insert(self.parts.upper, {body = partB, shape = partS, fixture = partF})
    
    --Joint
    local joint = love.physics.newRevoluteJoint(lastpartB, partB, 0, -height*(i-1)-offset, false)
    table.insert(self.revoluteJoints.up, joint)
    --Limits
    joint:setLimitsEnabled(true)
    joint:setLimits(-Bow.maxBend, -Bow.minBend)
    --Motor
    joint:setMotorEnabled(true)
    joint:setMotorSpeed(500)
    joint:setMaxMotorTorque(0)
    --Set the last part so the next division will be attached to this one
    lastpartB = partB
  end
  local end2 = lastpartB
  
  --The line
  local lineHandleB = love.physics.newBody(pWorld, -6, 0, "dynamic")
  local lineHandleS = love.physics.newRectangleShape(1, 7)
  local lineHandleF = love.physics.newFixture(lineHandleB, lineHandleS, 1)
  lineHandleB:setBullet(true)
  self.parts.lineHandle = {body = lineHandleB, shape = lineHandleS, fixture = lineHandleF}
  lineHandleB:setFixedRotation(true)
  
  local x1, y1 = lineHandleB:getWorldPoint(0, 3.25)
  local x2, y2 = end1:getWorldPoint(0, height/2)
  local lineJoint = love.physics.newDistanceJoint(lineHandleB, end1, x1, y1, x2, y2, height, false)
  lineJoint:setLength(Bow.lineLength/2)
  lineJoint:setFrequency(Bow.lineTension)
  lineJoint:setDampingRatio(.7)
  
  local x1, y1 = lineHandleB:getWorldPoint(0, -3.25)
  local x2, y2 = end2:getWorldPoint(0, -height/2)
  local lineJoint = love.physics.newDistanceJoint(lineHandleB, end2, x1, y1, x2, y2, height, false)
  lineJoint:setLength(Bow.lineLength/2)
  lineJoint:setFrequency(Bow.lineTension)
  lineJoint:setDampingRatio(.7)
end

function Bow:update(dt)
  local mul = 0
  if love.keyboard.isDown("d") then
    mul = mul + 1
  end
  
  if love.keyboard.isDown("a") then
    mul = mul - 1
  end
  self.parts.handle.body:setAngularVelocity(1*mul)
  
  --Update the tension on each division
  for n, joint in pairs(self.revoluteJoints.down) do
    local deltaAngle = joint:getJointAngle()-joint:getLowerLimit()
    local k = (joint:getUpperLimit()-joint:getLowerLimit())^2 * Bow.armTension
    local tension = deltaAngle*k
    joint:setMaxMotorTorque(tension)
  end
  for n, joint in pairs(self.revoluteJoints.up) do
    local deltaAngle = joint:getJointAngle()-joint:getUpperLimit()
    local k = (joint:getUpperLimit()-joint:getLowerLimit())^2 * Bow.armTension
    local tension = deltaAngle*k
    joint:setMaxMotorTorque(tension)
  end
  
  local maximumForce = Bow.pullStrength
  if love.mouse.isDown("1") then
    self.mousejoint:setTarget(d__getMousePosition())
  end
  if self.arrow then
    self.arrow.body:setPosition(self.parts.lineHandle.body:getWorldPoint(21.2, 0))
    self.arrow.body:setAngle(self.parts.handle.body:getAngle())
    self.arrow.body:setLinearVelocity(0, 0)
  end
  
  self.parts.lineHandle.body:setAngle(self.parts.handle.body:getAngle())
end

function Bow:draw()
  love.graphics.push()
  love.graphics.translate(d__getMousePosition())
  local angle = self.parts.handle.body:getAngle()
  love.graphics.line(0, 0, math.cos(angle)*50, math.sin(angle)*50)
  love.graphics.pop()
end

function Bow:keypressed(key)
  if key == "s" and love.mouse.isDown(1) and not self.arrow then
    local body = self.parts.lineHandle.body
    local ax, ay = body:getWorldPoint(13, 0)
    self.arrow = Arrow:new(ax, ay, 20, 1)
    Updater:add("bow", self.arrow)
  end
end

function Bow:mousepressed(x, y, button)
  if button == 1 then
    local body = self.parts.lineHandle.body
    local mx, my = d__getMousePosition()
    self.mousejoint = love.physics.newMouseJoint(body, mx, my)
    self.mousejoint:setMaxForce(600)
  end
end

function Bow:mousereleased(x, y, button)
  if button == 1 then
    self.mousejoint:destroy()
    self.mousejoint = nil
    if self.arrow then
      self.arrow = nil
    end
  end
end

return Bow